# Conclusion

I was not able to rebuild provided UnrealScript code. Lack of verbose feedback from 
build tools in UDK ecosystem (in particular, UDK.exe make) in case of problems has
exhausted me so I stopped researching on this game. Same as I stopped playing it.

# Deploy
TODO:
- Custom maps
- Proper shutdown with game status control
- Solve logging with custom /dev/null: `mknod /path/to/black/hole c 1 3`
- Apply webadmin patch
- Establish updates using <https://superuser.com/questions/1727148/check-if-steam-game-requires-an-update-via-command-line>
- Change as nothing as possible
- Took passwords to docker secrets/env and apply only at deploy time, use cli tool *crudini*?
    - Change current password as it now in git history of this repo
- Write game version on config update commit

# Reverse engineering & Tricks

TODO: 
- Fix chat encoding
- ~~Investigate `LogInternal` slow (buffering?)~~
- Proper shutdown with game status control
- increase people amount
    - play with config
    - inspect the mutator
- fix discord presence (at least for local client)
- Investigate balance_tweaks.bin
- Investigate using domain name in JoinString
- Investigate server to client RCE using `client reliable function`

## Console commands execution
From `Core/Classes/Actor.uc`:
```
// Execute a console command in the context of the current level and game engine.
native function string ConsoleCommand(string Command, optional bool bWriteToLog = true);
```

So it appears that for some console commands execution logic located in c++ code and
for some of them is defined in unrealscript as *exec function*s.

## UDK Versions

- 12791 - Latest available (<https://archive.org/download/udkinstaller/UDKInstall-2015-02.exe>); Tried to compile against it
- 10897 - KF2 is compiled against this version

## The GameInfo class
Might want to look to it for players amount increase.
```
The GameInfo class
The GameInfo class implements the game rules. A server (both dedicated and single-player) has one GameInfo subclass, accessible in UnrealScript as WorldInfo.Game. For each game type in Unreal, there is a special GameInfo subclass. For example, some existing classes are UTGame, UTDeathmatch, UTTeamGame.

A client in a network game does not have a GameInfo. That is, WorldInfo.Game == None on the client side. Clients should not be expected to have a GameInfo because the server implements all of the game play rules, and the generality of the code calls for the client not knowing what the game rules are.

GameInfo implements a broad set of functionality, such as recognizing players coming and going, assigning credit for kills, determining whether weapons should respawn, and so on. Here, we will only look at the GameInfo functions which are directly related to network programming. 
```
src: <https://docs.unrealengine.com/udk/Three/NetworkingOverview.html>

## Logging
From `Core/Classes/Object.uc`:
```
//
// Logging.
//
/**
 * Writes a message to the log.  This function should never be called directly - use the `log macro instead, which has the following signature:
 *
 * log( coerce string Msg, optional bool bCondition=true, optional name LogTag='ScriptLog' );
 *
 * @param   Msg             the string to print to the log
 * @param   bCondition      if specified, the message is only printed to the log if this condition is satisfied.
 * @param   LogTag          if specified, the message will be prepended with this tag in the log file
 *
 */
native(231) final static `{prevent_direct_calls} function LogInternal( coerce string S, optional name Tag );

/**
 * Same as calling LogInternal(SomeMsg, 'Warning');  This function should never be called directly - use the `warn macro instead, which has the following signature:
 *
 * warn( coerce string Msg, optional bool bCondition=true );
 */
native(232) final static `{prevent_direct_calls} function WarnInternal( coerce string S );
```

This function being called on every log record, located in KFGameServ binary file: `0x0000000000e43210  UObject::execLogInternal(FFrame&, void*)`

### Buffering solution

Despite it isn't state anywhere in documentation (or I haven't succeed in finding it), the 
logging system of unreal engine uses buffering.

It appears it's enough to add `-FORCELOGFLUSH` argument to avoid buffering.
src: <https://docs.unrealengine.com/udk/Three/CommandLineArguments.html>

Also check `-NOWRITE` - Disable output to log. I hope it disables log file writing. **Now it appears to don't work.**

It seems under windows there is no buffering but not sure, testing runs have not exceeded 30KB log size per run.

### Logging to file

Initially I wanted also to deal with log files rotation as I don't need them when running server in container.
But on time of tests (8 Jun 2023) server very unreliable writes logs to file at all. For around one hundred runs I conducted,
`Launch.log` file appeared only once.

Tested on windows and it appears to write logs to file just fine. Logs rotation also happens just fine.

## Shutdown
Shutdown with exit code 0 can be achieved using `quit` console command in webadmin.
```
curl 'http://192.168.1.25:8081/ServerAdmin/console' -X POST -H 'Content-Type: application/x-www-form-urlencoded' -H 'Cookie: authcred="YWRtaW4KQWRtaW5UaGVCZXN0"' 'command=quit'
```

Also investigate first line of logs: `Shutdown handler: initalize.`

WebAdmin console command stack:

- QHCurrent.handleQuery
- QHCurrent.handleConsole
- QHCurrent.adminCmdHandler.execute

Tested on windows and it handles shutdown by ctrl+c just fine, there a bunch of log lines regarding shutdown
procedure and then shutdowns (not sure if fully gracefull, haven't checked exit code), i.e.:

```
[0023.28] Log: 
[0023.28] Log: Success - 0 error(s), 0 warning(s)
[0023.28] Log: 
Execution of commandlet took:  16.87 seconds
[0023.29] ScriptLog: AccessControl CLEAN UP
[0023.33] Warning: Warning, Failed to load 'Class None.': Failed to find object 'Class None.'
[0023.33] Warning: Warning, Failed to find object 'Class None.'
[0023.34] Warning: Warning, Failed to load 'Class None.': Failed to find object 'Class None.'
[0023.34] Warning: Warning, Failed to find object 'Class None.'
[0023.84] Exit: Preparing to exit.
[0024.56] Exit: Game engine shut down
[0024.59] Exit: TcpNetDriver_0 shut down
[0024.69] Exit: Object subsystem successfully closed.
[0024.76] Exit: Exiting.
[0024.77] Log: Log file closed, 07/28/23 20:40:16
```

### Signals listening
```
SigIgn: TRAP(5) ALRM(14)
SigCgt: ILL(4) ABRT(6) BUS(7) FPE(8) SEGV(11) (33)
```
None of them invokes successful (with exit code 0) shutdown

## Get game version

`GAMEVER` console command returns info I don't see as applicapable to use.  
Steamquery's `get_info` function returns proper version. The same version can
be found in window's title of a desktop game client.

## Packages GUID manipulations
Get GUID of a package in python:
```python
# It's placed from 69 (dec) byte and takes 16 (dec) bytes in size 
import uuid
with open("WebAdmin.u", "br") as file:
    file.seek(69)
    r = file.read(16)

print(uuid.UUID(bytes_le=r))
```

Write GUID:
```python
import uuid
u = uuid.UUID('df835886-7616-4d01-b31f-f693ec9ffa71')
with open("WebAdmin.u", "r+b") as file:
    file.seek(69)
    file.write(u.bytes_le)
```

It's enough to set GUID of an original package to make game consider it as the same package.
GUIDs appears to change every update.
In case of WebAdmin.u, on mismatch the game tries to download it from Redirect server.

## Parse maps cycle in python
```python
maps.split('=')[2].split(')')[0].replace('(', '[') + ']'
```

## Language fix to try
```1. Кракозябры в админке:
Почему: сервер читает значение переменной LANG и пытается в мультиязычность (но не может).
Лечение: Перед запуском сервера установить LANG в английский, например так:
export LANG=en_US.utf8
```
src: https://steamcommunity.com/sharedfiles/filedetails/?id=1298957956

This method appears only to work with fields *GAME TYPE*, *DIFFICULTY*, *PERK* in webadmin.

## Language & encoding RE

Player to webadmin module still gives:
```
// `log("Message "$entry.message,,'WebAdmin');
[0092.35] WebAdmin: Message 83@>:1
```
```
// `log("BasicWebAdminUser.ReceiveMessage: " $Msg,,, 'WebAdmin');
BasicWebAdminUser.ReceiveMessage: B5AB B5AB
```

`IPDrv.WebResponse.CharSet` appears to not work

Sending message from web-admin call stack from server's view:

- WebAdmin.u: WebAdmin.Query
- WebAdmin.u: QHCurrentKF.handleQuery
- WebAdmin.u: QHCurrent.handleCurrentChatData
- WebAdmin.u: QHCurrent.BroadcastMessage
- Engine.u: webadmin.WorldInfo.Game.BroadcastHandler.BroadcastText
- Engine.u: PlayerController.TeamMessage // `reliable client simulated`
    - Engine.u: HUD.Message  // It appears to add console message
    - Engine.u: Console.OutputText // It's definitely the console

From a player to web admin call stack from server's view: 
- Engine.u: PlayerController.ServerSay  // `exec`; Client
- KFGame.u: KFPlayerController.ServerSay  // `unreliable server`; Server
- Engine.u: PlayerController.ServerSay
- Engine.u: GameInfo.Broadcast
- Engine.u: BroadcastHandler.Broadcast
- Engine.u: BroadcastHandler.BroadcastText
- WebAdmin.u: MessagingSpectator.TeamMessage
- WebAdmin.u: BasicWebAdminUser.ReceiveMessage


Team message are sent to client for execution so it actually happens on client with server's provided arguments.

On windows it appears to give another set of trash symbols (*charset*) (tested with applied utf-8 patch to frontend):

- From webadmin to player gives the same *charset* for player as if server under linux.
- From webadmin to webadmin gives windows specific *charset*. It also means webadmin <-> webadmin communication
is broken under windows despite it works under linux with applied utf-8 patch to frontend.
- From player to webadmin gives windows specific *charset*.

Tests on windows was conducted with russian and english system languages.

## Modules patching
How to log with substition:
```
`log("Message "$entry.message,,'WebAdmin');
// Maps to
LogInternal("Message " $ Entry.Message, 'WebAdmin');
```

# Patch to webadmin
This patch should be applied to `KFGame/Web/ServerAdmin`

```diff
diff --git a/current_players_row.inc b/current_players_row.inc
index 9851873..f1cad82 100755
--- a/current_players_row.inc
+++ b/current_players_row.inc
@@ -2,9 +2,17 @@
   <td style="background: <%player.teamcolor%>; color: <%player.teamcolor%>;"><%player.teamid%>&#160;</td>
   <td><%player.name%></td>
   <td class="right"><%player.ping%></td>
-  <td><%player.ip%></td>
+  <td><a href="https://kf2.demb.uk:444/ServerAdmin/systemd_api/whois?ip=<%player.ip%>">^M
+    <div style="height:100%;width:100%">^M
+    <%player.ip%>^M
+  </div>^M
+</a></td>^M
   <td><%player.uniqueid%></td>
-  <td><%player.steamid%></td>
+  <td><a href="https://steamcommunity.com/profiles/<%player.steamid%>">^M
+  <div style="height:100%;width:100%">^M
+    <%player.steamid%>^M
+  </div>^M
+</a></td>^M
   <td><%player.steamname%></td>
   <td class="center"><%player.admin%></td>
   <td class="center"><%player.spectator%></td>
diff --git a/header_base.inc b/header_base.inc
index 217a460..712fd3c 100755
--- a/header_base.inc
+++ b/header_base.inc
@@ -12,7 +12,7 @@
        var pageUri = '<%page.fulluri%>';
        </script>
        <script type="text/javascript" src="/images/jquery.js?gzip"></script>
-       <link href='http://fonts.googleapis.com/css?family=Open+Sans:400,400italic,600,700,700italic,600italic' rel='stylesheet' type='text/css'>
+       <link href='https://fonts.googleapis.com/css?family=Open+Sans:400,400italic,600,700,700italic,600italic' rel='stylesheet' type='text/css'>^M
        <%html.headers%>
 </head>
 <body class="<%page.css.class%>">
diff --git a/navigation.inc b/navigation.inc
index a374dd1..213b986 100755
--- a/navigation.inc
+++ b/navigation.inc
@@ -2,4 +2,5 @@
 <div id="menu">
 <h2>Navigation</h2>
 <%navigation.menu%>
+<li class="no-submenu"><a href="/ServerAdmin/systemd_api/restart" title="Restart server"><span>Restart server</span></a></li>^M
 </div>
```

This patch allows to load pages in UTF-8 by default so chat between webadmins on linux server is fine 

```diff
diff --git a/header_base.inc b/header_base.inc
index 217a460..e3d29ff 100755
--- a/header_base.inc
+++ b/header_base.inc
@@ -13,6 +13,7 @@
        </script>
        <script type="text/javascript" src="/images/jquery.js?gzip"></script>
        <link href='http://fonts.googleapis.com/css?family=Open+Sans:400,400italic,600,700,700italic,600italic' rel='stylesheet' type='text/css'>
+       <meta charset="UTF-8">
        <%html.headers%>
 </head>
 <body class="<%page.css.class%>">
```

# References
## Information
- https://wiki.killingfloor2.com/index.php?title=Dedicated_Server_(Killing_Floor_2)
- https://docs.unrealengine.com/udk/Three/CommandletList.html
- https://docs.unrealengine.com/udk/Three/PackagesAndNetworking.html
- https://docs.unrealengine.com/udk/Three/CharacterEncoding.html
- https://docs.unrealengine.com/udk/Three/UnrealScriptFunctions.html
- https://tripwireinteractive.atlassian.net/wiki/spaces/KF2SW/pages/26247172/KF2+Code+Modding+How-to
- https://steamcommunity.com/sharedfiles/filedetails/?id=1298957956
- https://docs.unrealengine.com/udk/Three/ConsoleCommands.html
- https://docs.unrealengine.com/udk/Three/ConfigurationFiles.html
- `killingfloor2\Development\Src` - Source code of some packages
- https://docs.unrealengine.com/udk/Three/UnrealScriptDefaultProperties.html
- https://docs.unrealengine.com/udk/Three/NetworkingOverview.html
- https://docs.unrealengine.com/udk/Three/NetworkProfiler.html
- https://docs.unrealengine.com/udk/Three/UT3Mods.html
- https://docs.unrealengine.com/udk/Three/CommandLineArguments.html
- https://docs.unrealengine.com/udk/Three/UnrealScriptFoundations.html

## Tools
- https://archive.org/download/udkinstall-2015-02
- https://eliotvu.com/portfolio/view/21/ue-explorer

# Draft
```
SETNOPEC GameInfo MaxPlayersAllowed 40
SETNOPEC GameInfo MaxPlayers 40


GET GameInfo MaxPlayersAllowed
GET GameInfo MaxPlayers

SETNOPEC KFAISpawnManager bLogAISpawning false



We want to change
WorldInfo.Game.MaxPlayersAllowed
WorldInfo.Game.MaxPlayers




АБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯ
абвгдеёжзийклмнопрстуфхцчшщъыьэюя

+32 (dec)


АБВГД
<span class="message"></span>


абвгд
<span class="message">012334</span>


я - O - 4F
Я - / - 2F
```